@@ -270,6 +270,16 @@ class DensityFlow(nn.Module):
                     allow_broadcast=self.allow_broadcast,
                     use_cuda=self.use_cuda,
                 )
+        if self.loss_func == 'negbinomial':
+            self.decoder_dispersion = MLP(
+                    [self.latent_dim] + self.decoder_hidden_layers + [self.input_size],
+                    activation=activate_fct,
+                    output_activation=Exp,
+                    post_layer_fct=post_layer_fct,
+                    post_act_fct=post_act_fct,
+                    allow_broadcast=self.allow_broadcast,
+                    use_cuda=self.use_cuda,
+                )
 
         if self.latent_dist == 'studentt':
             self.codebook = MLP(
@@ -398,13 +408,13 @@ class DensityFlow(nn.Module):
                     rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
-                logits = (mu.log()-dispersion.log()).clamp(min=-10, max=10)
+                logits = (mu.log()-dispersion.log()).clamp(min=-15, max=15)
+                #logits = (dispersion.log()-mu.log()).clamp(min=-15, max=15)
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, 
                                                                                          logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
                 else:
-                    pyro.sample('x', dist.NegativeBinomial(total_count=dispersion, 
-                                                           logits=logits).to_event(1), obs=xs)
+                    pyro.sample('x', dist.NegativeBinomial(total_count=dispersion, logits=logits).to_event(1), obs=xs)
             elif self.loss_func == 'poisson':
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
@@ -434,9 +444,9 @@ class DensityFlow(nn.Module):
         batch_size = xs.size(0)
         self.options = dict(dtype=xs.dtype, device=xs.device)
         
-        if self.loss_func=='negbinomial':
-            dispersion = pyro.param("dispersion", self.dispersion *
-                                            xs.new_ones(self.input_size), constraint=constraints.positive)
+        #if self.loss_func=='negbinomial':
+        #    dispersion = pyro.param("dispersion", self.dispersion *
+        #                                    xs.new_ones(self.input_size), constraint=constraints.positive)
             
         if self.use_zeroinflate:
             gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
@@ -480,6 +490,7 @@ class DensityFlow(nn.Module):
                 log_theta = log_mu
             elif self.loss_func == 'negbinomial':
                 mu = log_mu.exp()
+                dispersion = self.decoder_dispersion(zs)
             else:
                 rate = log_mu.exp()
                 theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
@@ -487,7 +498,8 @@ class DensityFlow(nn.Module):
                     rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
-                logits = (mu.log()-dispersion.log()).clamp(min=-10, max=10)
+                #logits = (mu.log()-dispersion.log()).clamp(min=-15, max=15)
+                logits = (dispersion.log()-mu.log()).clamp(min=-15, max=15)
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
                 else:
@@ -586,7 +598,8 @@ class DensityFlow(nn.Module):
                     rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
-                logits = (mu.log()-dispersion.log()).clamp(min=-10, max=10)
+                logits = (mu.log()-dispersion.log()).clamp(min=-15, max=15)
+                #logits = (dispersion.log()-mu.log()).clamp(min=-15, max=15)
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, 
                                                                                          logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
@@ -696,7 +709,8 @@ class DensityFlow(nn.Module):
                     rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
-                logits = (mu.log()-dispersion.log()).clamp(min=-10, max=10)
+                logits = (mu.log()-dispersion.log()).clamp(min=-15, max=15)
+                #logits = (dispersion.log()-mu.log()).clamp(min=-15, max=15)
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, 
                                                                                          logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
@@ -765,7 +779,7 @@ class DensityFlow(nn.Module):
         dzs = self._total_effects(basal, us)
         return basal + dzs 
     
-    def get_complete_embedding(self, xs, us, batch_size:int=1024):
+    def get_complete_embedding(self, xs, us, batch_size:int=1024, show_progress=True):
         xs = self.preprocess(xs)
         xs = convert_to_tensor(xs, device=self.get_device())
         us = convert_to_tensor(us, device=self.get_device())
@@ -773,7 +787,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         Z = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for X_batch, U_batch, _ in dataloader:
                 zns = self._get_complete_embedding(X_batch, U_batch)
                 Z.append(tensor_to_numpy(zns))
@@ -788,7 +802,8 @@ class DensityFlow(nn.Module):
     
     def get_basal_embedding(self, 
                              xs, 
-                             batch_size: int = 1024):
+                             batch_size: int = 1024,
+                             show_progress=True):
         """
         Return cells' basal latent representations
 
@@ -809,7 +824,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         Z = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for X_batch, _ in dataloader:
                 zns,_ = self._get_basal_embedding(X_batch)
                 Z.append(tensor_to_numpy(zns))
@@ -827,14 +842,14 @@ class DensityFlow(nn.Module):
             alpha = self.encoder_n(zns)
         return alpha
     
-    def code(self, xs, batch_size=1024):
+    def code(self, xs, batch_size=1024, show_progress=True):
         xs = self.preprocess(xs)
         xs = convert_to_tensor(xs, device=self.get_device())
         dataset = CustomDataset(xs)
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         A = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for X_batch, _ in dataloader:
                 a = self._code(X_batch)
                 A.append(tensor_to_numpy(a))
@@ -848,7 +863,7 @@ class DensityFlow(nn.Module):
         alpha = self.softmax(alpha)
         return alpha
     
-    def soft_assignments(self, xs, batch_size=1024):
+    def soft_assignments(self, xs, batch_size=1024, show_progress=True):
         """
         Map cells to metacells and return the probabilistic values of metacell assignments
         """
@@ -858,7 +873,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         A = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for X_batch, _ in dataloader:
                 a = self._soft_assignments(X_batch)
                 A.append(tensor_to_numpy(a))
@@ -873,7 +888,7 @@ class DensityFlow(nn.Module):
         ns = torch.zeros_like(alpha).scatter_(1, ind, 1.0)
         return ns
     
-    def hard_assignments(self, xs, batch_size=1024):
+    def hard_assignments(self, xs, batch_size=1024, show_progress=True):
         """
         Map cells to metacells and return the assigned metacell identities.
         """
@@ -883,7 +898,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         A = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for X_batch, _ in dataloader:
                 a = self._hard_assignments(X_batch)
                 A.append(tensor_to_numpy(a))
@@ -892,22 +907,22 @@ class DensityFlow(nn.Module):
         A = np.concatenate(A)
         return A
     
-    def predict(self, xs, us, perturbs_predict:list, perturbs_reference:list, library_sizes=None):
+    def predict(self, xs, us, perturbs_predict:list, perturbs_reference:list, library_sizes=None, show_progress=True):
         perturbs_reference = np.array(perturbs_reference)
         
         # basal embedding
-        zs = self.get_basal_embedding(xs)
+        zs = self.get_basal_embedding(xs, show_progress=show_progress)
         for pert in perturbs_predict:
             pert_idx = int(np.where(perturbs_reference==pert)[0])
             us_i = us[:,pert_idx].reshape(-1,1)
             
             # factor effect of xs 
-            dzs0 = self.get_cell_shift(zs, perturb_idx=pert_idx, perturb_us=us_i)
+            dzs0 = self.get_cell_shift(zs, perturb_idx=pert_idx, perturb_us=us_i, show_progress=show_progress)
             
             # perturbation effect 
             ps = np.ones_like(us_i)
             if np.sum(np.abs(ps-us_i))>=1:
-                dzs = self.get_cell_shift(zs, perturb_idx=pert_idx, perturb_us=ps)          
+                dzs = self.get_cell_shift(zs, perturb_idx=pert_idx, perturb_us=ps, show_progress=show_progress)          
                 zs = zs + dzs0 + dzs
             else:
                 zs = zs + dzs0
@@ -920,8 +935,8 @@ class DensityFlow(nn.Module):
         elif len(library_sizes.shape)==1:
             library_sizes = library_sizes.reshape(-1,1)
             
-        counts = self.get_counts(zs, library_sizes=library_sizes)
-        log_mu = self.get_log_mu(zs)
+        counts = self.get_counts(zs, library_sizes=library_sizes, show_progress=show_progress)
+        log_mu = self.get_log_mu(zs, show_progress=show_progress)
         
         return counts, log_mu
     
@@ -946,7 +961,8 @@ class DensityFlow(nn.Module):
                              zs, 
                              perturb_idx,
                              perturb_us,
-                             batch_size: int = 1024):
+                             batch_size: int = 1024,
+                             show_progress=True):
         """
         Return cells' changes in the latent space induced by specific perturbation of a factor
 
@@ -958,7 +974,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         Z = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for Z_batch, P_batch, _ in dataloader:
                 zns = self._cell_shift(Z_batch, perturb_idx, P_batch)
                 Z.append(tensor_to_numpy(zns))
@@ -970,7 +986,7 @@ class DensityFlow(nn.Module):
     def _log_mu(self, zs):
         return self.decoder_log_mu(zs)
     
-    def get_log_mu(self, zs, batch_size: int = 1024):
+    def get_log_mu(self, zs, batch_size: int = 1024, show_progress=True):
         """
         Return cells' changes in the feature space induced by specific perturbation of a factor
 
@@ -980,7 +996,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
 
         R = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for Z_batch, _ in dataloader:
                 r = self._log_mu(Z_batch)
                 R.append(tensor_to_numpy(r))
@@ -1002,7 +1018,8 @@ class DensityFlow(nn.Module):
         return counts
     
     def get_counts(self, zs, library_sizes,
-                        batch_size: int = 1024):
+                        batch_size: int = 1024,
+                        show_progress=True):
 
         zs = convert_to_tensor(zs, device=self.get_device())
         
@@ -1016,7 +1033,7 @@ class DensityFlow(nn.Module):
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
         
         E = []
-        with tqdm(total=len(dataloader), desc='', unit='batch') as pbar:
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
             for Z_batch, L_batch, _ in dataloader:
                 log_mu = self._log_mu(Z_batch)
                 counts = self._count(log_mu, L_batch)
@@ -1031,7 +1048,7 @@ class DensityFlow(nn.Module):
             ad = sc.AnnData(xs)
             binarize(ad, threshold=threshold)
             xs = ad.X.copy()
-        elif self.loss_func == 'poisson':
+        else:
             xs = np.round(xs)
             
         if sparse.issparse(xs):
@@ -1051,7 +1068,8 @@ class DensityFlow(nn.Module):
             decay_rate: float = 0.9,
             config_enum: str = 'parallel',
             threshold: int = 0,
-            use_jax: bool = True):
+            use_jax: bool = True,
+            show_progress=True):
         """
         Train the DensityFlow model.
 
@@ -1131,7 +1149,7 @@ class DensityFlow(nn.Module):
         losses = [loss_basic]
         num_losses = len(losses)
 
-        with tqdm(total=num_epochs, desc='Training', unit='epoch') as pbar:
+        with tqdm(total=num_epochs, disable=not show_progress, desc='Training', unit='epoch') as pbar:
             for epoch in range(num_epochs):
                 epoch_losses = [0.0] * num_losses
                 for batch_x, batch_u, batch_y, batch_z, _ in dataloader:
