@@ -1,6 +1,6 @@
 from .SURE import SURE
 from .DensityFlow import DensityFlow
-from .DensityFlowLinear import DensityFlowLinear
+from .DensityFlowNCB import DensityFlowNCB
 from .PerturbE import PerturbE
 from .TranscriptomeDecoder import TranscriptomeDecoder
 from .SimpleTranscriptomeDecoder import SimpleTranscriptomeDecoder
@@ -12,7 +12,7 @@ from . import utils
 from . import codebook
 from . import SURE
 from . import DensityFlow
-from . import DensityFlowLinear
+from . import DensityFlowNCB
 from . import atac
 from . import flow 
 from . import perturb
@@ -23,6 +23,7 @@ from . import EfficientTranscriptomeDecoder
 from . import VirtualCellDecoder
 from . import PerturbationAwareDecoder
 
-__all__ = ['SURE', 'DensityFlow', 'DensityFlowLinear', 'PerturbE', 'TranscriptomeDecoder', 'SimpleTranscriptomeDecoder',
+__all__ = ['SURE', 'DensityFlow', 'DensityFlowNCB', 'PerturbE', 
+           'TranscriptomeDecoder', 'SimpleTranscriptomeDecoder',
            'EfficientTranscriptomeDecoder', 'VirtualCellDecoder', 'PerturbationAwareDecoder', 
            'flow', 'perturb', 'atac', 'utils', 'codebook']
\ No newline at end of file
